
TitleBelow = 0
TitleAbove = 1

NoPosition = 0
AxisStart = 1
AxisEnd = 2
AxisMiddle = 4

yLeft = 0
yRight = 1
xBottom = 2
xTop = 3
UserAxis = 100

XAxes = [xBottom, xTop]
YAxes = [yLeft, yRight]
CartesianAxes = XAxes + YAxes

ProbabilitiesZValue = -20
SelectionZValue = -10
AxisZValue = -2
PlotZValue = 0
HighlightZValue = 5
LegendZValue = 10

NOTHING = 0
ZOOMING = 1
SELECT = 2
SELECT_POLYGON = 3
PANNING = 4
SELECT_RECTANGLE = SELECT
SELECT_RIGHTCLICK = SELECT

SELECTION_ADD = 0
SELECTION_REMOVE = 1
SELECTION_TOGGLE = 2
SELECTION_REPLACE = 3

# Used in 3D widgets
ROTATING = 5
