"""
=========
Visualize
=========

Widgets for data visualization.

"""

# Category description for the widget registry

NAME = "Visualize"

ID = "orange.widgets.visualize"

DESCRIPTION = "Widgets for data visualization."

BACKGROUND = "#FFB7B1"

ICON = "icons/Category-Visualize.svg"

PRIORITY = 2
